/* QNT - user.cpp - File buffer class / user callable functions. */


#define WIN32_LEAN_AND_MEAN
#include <ctype.h>
#include <limits.h>
#include <wtypes.h>
#include "buffer.h"
#include "clipb.h"
#include "config.h"
#include "display.h"


/***********************************************************************
    Text insertion:
***********************************************************************/

BOOL Buffer::litchar( char c )
{
    literal( c );
    changed = TRUE;
    return TRUE;
}


BOOL Buffer::enter()
{
    for( int i = 0; i < enterlen[ent]; i++ )
        startchars = new CharList( enterstr[ent][i], -2, startchars );
    int size = startchars->count();
    char *buf = new char[size];
    startchars->putrev( buf + size - 1 );
    delete startchars;
    startlines = new LineList( size - enterlen[ent], size, buf, startlines );
    startchars = new CharList;
    row++;
    rows++;
    col = 0;
    changed = TRUE;
    return TRUE;
}


/***********************************************************************
    Edit commands operating on single line:
***********************************************************************/

BOOL Buffer::charleft()
{
    char ch = startchars->getch();
    signed char tab = startchars->gettab();
    if( tab != -1 )
    {
        if( tabsize == 0 || ch != '\t' )
            col--;
        else
            col -= tab;
        startchars = startchars->cdr();
        endchars = new CharList( ch, endchars );
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::charright()
{
    if( endchars->gettab() >= 0 )
    {
        literal( endchars->getch() );
        endchars = endchars->cdr();
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::linebeg()
{
    for( int i = -1; charleft(); i++ );
    return !!i;
}


BOOL Buffer::lineend()
{
    for( int i = -1; charright(); i++ );
    return !!i;
}


BOOL Buffer::wordleft()
{
    BOOL prevletter = FALSE;
    BOOL letter;
    while( letter = !!isalnum( startchars->getch() ), letter || !prevletter )
    {
        if( !charleft() )
            return FALSE;
        prevletter = letter;
    }
    return TRUE;
}


BOOL Buffer::wordright()
{
    BOOL prevletter = TRUE;
    BOOL letter;
    while( letter = !!isalnum( endchars->getch() ), !letter || prevletter )
    {
        if( !charright() )
            return FALSE;
        prevletter = letter;
    }
    return TRUE;
}


BOOL Buffer::delcharleft()
{
    if( charleft() )
    {
        delcharright();
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::delcharright()
{
    if( endchars->gettab() >= 0 )
    {
        endchars = endchars->cdr();
        changed = TRUE;
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::dellinebeg()
{
    for( int i = -1; delcharleft(); i++ );
    return !!i;
}


BOOL Buffer::dellineend()
{
    for( int i = -1; delcharright(); i++ );
    return !!i;
}


BOOL Buffer::delwordleft()
{
    BOOL prevletter = FALSE;
    BOOL letter;
    while( letter = !!isalnum( startchars->getch() ), letter || !prevletter )
    {
        if( !delcharleft() )
            return FALSE;
        prevletter = letter;
    }
    return TRUE;
}


BOOL Buffer::delwordright()
{
    BOOL prevletter = TRUE;
    BOOL letter;
    while( letter = !!isalnum( endchars->getch() ), !letter || prevletter )
    {
        if( !delcharright() )
            return FALSE;
        prevletter = letter;
    }
    return TRUE;
}


/***********************************************************************
    Multiline versions of single line operations:
***********************************************************************/

BOOL Buffer::delcharleftwrap()
{
    if( !delcharleft() )
    {
        if( lineup() )
        {
            lineend();
            return delcharrightwrap();
        }
        else
            return FALSE;
    }
    else
        return TRUE;
}


BOOL Buffer::delcharrightwrap()
{
    if( !delcharright() )
    {
        int len = endlines->getlen();
        if( len )
        {
            int vislen = endlines->getvislen();
            char *buf = endlines->cdrbuf();
            endlines = endlines->cdr();
            delete endchars;
            endchars = new CharList;
            for( char *ptr = buf + len - 1; ptr >= buf + vislen; ptr-- )
                endchars = new CharList( *ptr, -2, endchars );
            for( ; ptr >= buf; ptr-- )
                endchars = new CharList( *ptr, endchars );
            changed = TRUE;
            return TRUE;
        }
        else
            return FALSE;
    }
    else
        return TRUE;
}


BOOL Buffer::delcharrightwrapsel()
{
    if( arow != row || acol != col )
        return delselect();
    else
        return delcharrightwrap();
}


/***********************************************************************
    Edit commands spanning multiple lines:
***********************************************************************/

BOOL Buffer::lineup()
{
    int len = startlines->getlen();
    if( len != 0 )
    {
        int vissize;
        int size;
        char *buf;
        exitline( vissize, size, buf );
        endlines = new LineList( vissize, size, buf, endlines );
        enterline( startlines->getvislen(), len, startlines->cdrbuf(), col );
        startlines = startlines->cdr();
        row--;
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::linedown()
{
    int len = endlines->getlen();
    if( len != 0 )
    {
        int vissize;
        int size;
        char *buf;
        exitline( vissize, size, buf );
        startlines = new LineList( vissize, size, buf, startlines );
        enterline( endlines->getvislen(), len, endlines->cdrbuf(), col );
        endlines = endlines->cdr();
        row++;
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::pageup()
{
    int i = display.getheight() - Config::getpageoverlap();
    int j;
    for( j = 0; j < i; j++ )
        if( !lineup() )
            break;
    drow = max( drow - j, 0 );
    invscreen = TRUE;
    return TRUE;
}


BOOL Buffer::pagedown()
{
    int i = display.getheight() - Config::getpageoverlap();
    int j;
    for( j = 0; j < i; j++ )
        if( !linedown() )
            break;
    drow = min( drow + j, rows );
    invscreen = TRUE;
    return TRUE;
}


BOOL Buffer::delline()
{
    delete startchars;
    delete endchars;
    int len = endlines->getlen();
    if( len != 0 )
    {
        enterline( endlines->getvislen(), len, endlines->cdrbuf(), col );
        endlines = endlines->cdr();
        rows--;
    }
    else
    {
        len = startlines->getlen();
        if( len != 0 )
        {
            enterline( startlines->getvislen(), len, startlines->cdrbuf(), col );
            startlines = startlines->cdr();
            rows--;
            row--;
        }
        else
        {
            startchars = new CharList;
            endchars = new CharList;
            col = 0;
        }
    }
    changed = TRUE;
    return TRUE;
}


/***********************************************************************
    Block commands:
***********************************************************************/

BOOL Buffer::delselect()
{
    if( row != arow || col != acol )
    {
        if( row < arow || (row == arow && col < acol) )
        {
            int r = row;
            int c = col;
            seekto( arow, acol );
            arow = r;
            acol = c;
        }
        if( row > arow )
        {
            delete startchars;
            startchars = new CharList;
            col = 0;
            while( row > arow + 1 )
            {
                startlines = startlines->cdr();
                row--;
            }
            delcharleftwrap();
        }
        while( col > acol )
            delcharleft();
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::paste()
{
    char *buf = Clipb::get();
    if( buf )
    {
        char *ptr = buf;
        while( *ptr )
        {
            if( *ptr == '\n' )
            {
                enter();
                if( ptr[1] == '\r' )
                    ptr++;
            }
            else if( *ptr == '\r' )
            {
                enter();
                if( ptr[1] == '\n' )
                    ptr++;
            }
            else
                litchar( *ptr );
            ptr++;
        }
        delete [] buf;
        return TRUE;
    }
    else
        return FALSE;
}


BOOL Buffer::copy()
{
    if( row != arow || col != acol )
    {
        if( row < arow || (row == arow && col < acol) )
        {
            int r = row;
            int c = col;
            seekto( arow, acol );
            arow = r;
            acol = c;
        }
        BOOL bool;
        int size = startchars->count();
        if( row > arow )
        {
            for( int i = row; i > arow + 1; i-- )
            {
                int vislen = startlines->getvislen();
                int len = startlines->getlen();
                char *lbuf = startlines->cdrbuf();
                startlines = startlines->cdr();
                size += vislen + Clipb::getenterlen();
                endlines = new LineList( vislen, len, lbuf, endlines );
            }
            char *buf = startlines->getbuf();
            char *cbuf = seekcol( buf, acol );
            int amnt = startlines->getvislen() - (cbuf - buf);
            size += amnt + Clipb::getenterlen();
            buf = new char[size + 1];
            buf[size] = 0;
            memcpy( buf, cbuf, amnt );
            cbuf = buf + amnt;
            memcpy( cbuf, Clipb::getenter(), Clipb::getenterlen() );
            cbuf += Clipb::getenterlen();
            while( i++ < row )
            {
                int vislen = endlines->getvislen();
                int len = endlines->getlen();
                char *lbuf = endlines->cdrbuf();
                endlines = endlines->cdr();
                memcpy( cbuf, lbuf, vislen );
                cbuf += vislen;
                memcpy( cbuf, Clipb::getenter(), Clipb::getenterlen() );
                cbuf += Clipb::getenterlen();
                startlines = new LineList( vislen, len, lbuf, startlines );
            }
            startchars->putrev( buf + size - 1 );
            bool = Clipb::put( buf );
            delete [] buf;
        }
        else
        {
            char *buf = new char[size + 1];
            buf[size] = 0;
            startchars->putrev( buf + size - 1 );
            bool = Clipb::put( seekcol( buf, acol ) );
            delete [] buf;
        }
        return bool;
    }
    else
        return FALSE;
}


BOOL Buffer::cut()
{
    if( copy() )
        return delselect();
    else
        return FALSE;
}
